Package: safetensors
Title: Safetensors File Format
Version: 0.2.1
Authors@R: c(
    person("Tomasz", "Kalinowski", , "tomasz@posit.co", role = c("ctb", "cre")),
    person("Daniel", "Falbel", , "dfalbel@gmail.com", role = "aut"),
    person("Sebastian", "Fischer", , "seb.fischer@tutamail.com", role = "ctb"),
    person(family = "Posit", role = c("cph"))
    )
Description: A file format for storing tensors that is secure (doesn't allow for
    code execution), fast and simple to implement. 'safetensors' also enables cross
    language and cross frameworks compatibility making it an ideal format for
    storing machine learning model weights.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: testthat (>= 3.0.0), torch (>= 0.11.0)
Config/testthat/edition: 3
Imports: cli, jsonlite, R6, rlang
URL: https://github.com/mlverse/safetensors,
        https://mlverse.github.io/safetensors/
BugReports: https://github.com/mlverse/safetensors/issues
NeedsCompilation: no
Packaged: 2026-04-27 11:19:51 UTC; dfalbel
Author: Tomasz Kalinowski [ctb, cre],
  Daniel Falbel [aut],
  Sebastian Fischer [ctb],
  Posit [cph]
Maintainer: Tomasz Kalinowski <tomasz@posit.co>
Repository: CRAN
Date/Publication: 2026-04-27 13:10:02 UTC
Built: R 4.4.3; ; 2026-05-01 12:56:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(safe_tensor_buffer,torch_tensor)
S3method(safe_tensor_meta,torch_tensor)
export(safe_load_file)
export(safe_save_file)
export(safe_serialize)
export(safe_tensor_buffer)
export(safe_tensor_meta)
export(safetensors)
export(safetensors_frameworks)
importFrom(R6,R6Class)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# safetensors

<!-- badges: start -->

[![R-CMD-check](https://github.com/mlverse/safetensors/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mlverse/safetensors/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

safetensors is a pure R implementation of the
[safetensors](https://github.com/safetensors/safetensors) file format
for both reading and writing. It currently supports the {torch} and
{pjrt} frameworks.

## Installation

safetensors can be installed from CRAN with:

``` r
install.packages("safetensors")
```

The development version of safetensors from
[GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("mlverse/safetensors")
```

## Example

Here’s an example of writing and reading safetensors files in
combination with torch:

``` r
library(torch)
library(safetensors)

tensors <- list(
  x = torch_randn(10, 10),
  y = torch_ones(10, 10)
)

str(tensors)
#> List of 2
#>  $ x:Float [1:10, 1:10]
#>  $ y:Float [1:10, 1:10]

tmp <- tempfile()
safe_save_file(tensors, tmp)

tensors <- safe_load_file(tmp, framework = "torch")
str(tensors)
#> List of 2
#>  $ x:Float [1:10, 1:10]
#>  $ y:Float [1:10, 1:10]
#>  - attr(*, "metadata")=List of 2
#>   ..$ x:List of 3
#>   .. ..$ shape       : int [1:2] 10 10
#>   .. ..$ dtype       : chr "F32"
#>   .. ..$ data_offsets: int [1:2] 0 400
#>   ..$ y:List of 3
#>   .. ..$ shape       : int [1:2] 10 10
#>   .. ..$ dtype       : chr "F32"
#>   .. ..$ data_offsets: int [1:2] 400 800
#>  - attr(*, "max_offset")= int 929
```

<p align="center">
  <picture>
    <source media="(prefers-color-scheme: dark)" srcset="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-dark.svg">
    <source media="(prefers-color-scheme: light)" srcset="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-light.svg">
    <img alt="Hugging Face Safetensors Library" src="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-light.svg" style="max-width: 100%;">
  </picture>
  <br/>
  <br/>
</p>

Python
[![Pypi](https://img.shields.io/pypi/v/safetensors.svg)](https://pypi.org/pypi/safetensors/)
[![Documentation](https://img.shields.io/website/http/huggingface.co/docs/safetensors/index.svg?label=docs)](https://huggingface.co/docs/safetensors/index)
[![Codecov](https://codecov.io/github/huggingface/safetensors/coverage.svg?branch=main)](https://codecov.io/gh/huggingface/safetensors)
[![Downloads](https://static.pepy.tech/badge/safetensors/month)](https://pepy.tech/project/safetensors)

Rust
[![Crates.io](https://img.shields.io/crates/v/safetensors.svg)](https://crates.io/crates/safetensors)
[![Documentation](https://docs.rs/safetensors/badge.svg)](https://docs.rs/safetensors/)
[![Codecov](https://codecov.io/github/huggingface/safetensors/coverage.svg?branch=main)](https://codecov.io/gh/huggingface/safetensors)
[![Dependency status](https://deps.rs/repo/github/huggingface/safetensors/status.svg?path=safetensors)](https://deps.rs/repo/github/huggingface/safetensors?path=safetensors)

# safetensors

## Safetensors

This repository implements a new simple format for storing tensors
safely (as opposed to pickle) and that is still fast (zero-copy).

### Installation
#### Pip

You can install safetensors via the pip manager:

```bash
pip install safetensors
```

#### From source

For the sources, you need Rust

```bash
# Install Rust
curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh
# Make sure it's up to date and using stable channel
rustup update
git clone https://github.com/huggingface/safetensors
cd safetensors/bindings/python
pip install setuptools_rust
pip install -e .
```

### Getting started

```python
import torch
from safetensors import safe_open
from safetensors.torch import save_file

tensors = {
   "weight1": torch.zeros((1024, 1024)),
   "weight2": torch.zeros((1024, 1024))
}
save_file(tensors, "model.safetensors")

tensors = {}
with safe_open("model.safetensors", framework="pt", device="cpu") as f:
   for key in f.keys():
       tensors[key] = f.get_tensor(key)
```

[Python documentation](https://huggingface.co/docs/safetensors/index)


### Format

- 8 bytes: `N`, an unsigned little-endian 64-bit integer, containing the size of the header
- N bytes: a JSON UTF-8 string representing the header.
  - The header data MUST begin with a `{` character (0x7B).
  - The header data MAY be trailing padded with whitespace (0x20).
  - The header is a dict like `{"TENSOR_NAME": {"dtype": "F16", "shape": [1, 16, 256], "data_offsets": [BEGIN, END]}, "NEXT_TENSOR_NAME": {...}, ...}`,
    - `data_offsets` point to the tensor data relative to the beginning of the byte buffer (i.e. not an absolute position in the file),
      with `BEGIN` as the starting offset and `END` as the one-past offset (so total tensor byte size = `END - BEGIN`).
  - A special key `__metadata__` is allowed to contain free form string-to-string map. Arbitrary JSON is not allowed, all values must be strings.
- Rest of the file: byte-buffer.

Notes:
 - Duplicate keys are disallowed. Not all parsers may respect this.
 - In general the subset of JSON is implicitly decided by `serde_json` for
   this library. Anything obscure might be modified at a later time, that odd ways
   to represent integer, newlines and escapes in utf-8 strings. This would only
   be done for safety concerns
 - Tensor values are not checked against, in particular NaN and +/-Inf could
   be in the file
 - Empty tensors (tensors with 1 dimension being 0) are allowed.
   They are not storing any data in the databuffer, yet retaining size in the header.
   They don't really bring a lot of values but are accepted since they are valid tensors
   from traditional tensor libraries perspective (torch, tensorflow, numpy, ..).
 - 0-rank Tensors (tensors with shape `[]`) are allowed, they are merely a scalar.
 - The byte buffer needs to be entirely indexed, and cannot contain holes. This prevents
   the creation of polyglot files.
 - Endianness: Little-endian.
   moment.
 - Order: 'C' or row-major.


### Yet another format ?

The main rationale for this crate is to remove the need to use
`pickle` on `PyTorch` which is used by default.
There are other formats out there used by machine learning and more general
formats.


Let's take a look at alternatives and why this format is deemed interesting.
This is my very personal and probably biased view:

| Format                  | Safe | Zero-copy | Lazy loading | No file size limit | Layout control | Flexibility | Bfloat16/Fp8
| ----------------------- | --- | --- | --- | --- | --- | --- | --- |
| pickle (PyTorch)        | ✗ | ✗ | ✗ | 🗸 | ✗ | 🗸 | 🗸 |
| H5 (Tensorflow)         | 🗸 | ✗ | 🗸 | 🗸 | ~ | ~ | ✗ |
| SavedModel (Tensorflow) | 🗸 | ✗ | ✗ | 🗸 | 🗸 | ✗ | 🗸 |
| MsgPack (flax)          | 🗸 | 🗸 | ✗ | 🗸 | ✗ | ✗ | 🗸 |
| Protobuf (ONNX)         | 🗸 | ✗ | ✗ | ✗ | ✗ | ✗ | 🗸 |
| Cap'n'Proto             | 🗸 | 🗸 | ~ | 🗸 | 🗸 | ~ | ✗ |
| Arrow                   | ? | ? | ? | ? | ? | ? | ✗ |
| Numpy (npy,npz)         | 🗸 | ? | ? | ✗ | 🗸 | ✗ | ✗ |
| pdparams (Paddle)       | ✗ | ✗ | ✗ | 🗸 | ✗ | 🗸 | 🗸 |
| SafeTensors             | 🗸 | 🗸 | 🗸 | 🗸 | 🗸 | ✗ | 🗸 |

- Safe: Can I use a file randomly downloaded and expect not to run arbitrary code ?
- Zero-copy: Does reading the file require more memory than the original file ?
- Lazy loading: Can I inspect the file without loading everything ? And loading only
  some tensors in it without scanning the whole file (distributed setting) ?
- Layout control: Lazy loading, is not necessarily enough since if the information about tensors is spread out in your file, then even if the information is lazily accessible you might have to access most of your file to read the available tensors (incurring many DISK -> RAM copies). Controlling the layout to keep fast access to single tensors is important.
- No file size limit: Is there a limit to the file size ?
- Flexibility: Can I save custom code in the format and be able to use it later with zero extra code ? (~ means we can store more than pure tensors, but no custom code)
- Bfloat16/Fp8: Does the format support native bfloat16/fp8 (meaning no weird workarounds are
  necessary)? This is becoming increasingly important in the ML world.


### Main oppositions

- Pickle: Unsafe, runs arbitrary code
- H5: Apparently now discouraged for TF/Keras. Seems like a great fit otherwise actually. Some classic use after free issues: <https://www.cvedetails.com/vulnerability-list/vendor_id-15991/product_id-35054/Hdfgroup-Hdf5.html>. On a very different level than pickle security-wise. Also 210k lines of code vs ~400 lines for this lib currently.
- SavedModel: Tensorflow specific (it contains TF graph information).
- MsgPack: No layout control to enable lazy loading (important for loading specific parts in distributed setting)
- Protobuf: Hard 2Go max file size limit
- Cap'n'proto: Float16 support is not present [link](https://capnproto.org/language.html#built-in-types) so using a manual wrapper over a byte-buffer would be necessary. Layout control seems possible but not trivial as buffers have limitations [link](https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize).
- Numpy (npz): No `bfloat16` support. Vulnerable to zip bombs (DOS). Not zero-copy.
- Arrow: No `bfloat16` support.

### Notes

- Zero-copy: No format is really zero-copy in ML, it needs to go from disk to RAM/GPU RAM (that takes time). On CPU, if the file is already in cache, then it can
  truly be zero-copy, whereas on GPU there is not such disk cache, so a copy is always required
  but you can bypass allocating all the tensors on CPU at any given point.
  SafeTensors is not zero-copy for the header. The choice of JSON is pretty arbitrary, but since deserialization is <<< of the time required to load the actual tensor data and is readable I went that way, (also space is <<< to the tensor data).

- Endianness: Little-endian. This can be modified later, but it feels really unnecessary at the
  moment.
- Order: 'C' or row-major. This seems to have won. We can add that information later if needed.
- Stride: No striding, all tensors need to be packed before being serialized. I have yet to see a case where it seems useful to have a strided tensor stored in serialized format.

### Benefits

Since we can invent a new format we can propose additional benefits:

- Prevent DOS attacks: We can craft the format in such a way that it's almost
  impossible to use malicious files to DOS attack a user. Currently, there's a limit
  on the size of the header of 100MB to prevent parsing extremely large JSON.
  Also when reading the file, there's a guarantee that addresses in the file
  do not overlap in any way, meaning when you're loading a file you should never
  exceed the size of the file in memory

- Faster load: PyTorch seems to be the fastest file to load out in the major
  ML formats. However, it does seem to have an extra copy on CPU, which we
  can bypass in this lib by using `torch.UntypedStorage.from_file`.
  Currently, CPU loading times are extremely fast with this lib compared to pickle.
  GPU loading times are as fast or faster than PyTorch equivalent.
  Loading first on CPU with memmapping with torch, and then moving all tensors to GPU seems
  to be faster too somehow (similar behavior in torch pickle)

- Lazy loading: in distributed (multi-node or multi-gpu) settings, it's nice to be able to
  load only part of the tensors on the various models. For
  [BLOOM](https://huggingface.co/bigscience/bloom) using this format enabled
  to load the model on 8 GPUs from 10mn with regular PyTorch weights down to 45s.
  This really speeds up feedbacks loops when developing on the model. For instance
  you don't have to have separate copies of the weights when changing the distribution
  strategy (for instance Pipeline Parallelism vs Tensor Parallelism).

License: Apache-2.0

<p align="center">
  <picture>
    <source media="(prefers-color-scheme: dark)" srcset="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-dark.svg">
    <source media="(prefers-color-scheme: light)" srcset="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-light.svg">
    <img alt="Hugging Face Safetensors Library" src="https://huggingface.co/datasets/safetensors/assets/raw/main/banner-light.svg" style="max-width: 100%;">
  </picture>
  <br/>
  <br/>
</p>

Python
[![Pypi](https://img.shields.io/pypi/v/safetensors.svg)](https://pypi.org/pypi/safetensors/)
[![Documentation](https://img.shields.io/website/http/huggingface.co/docs/safetensors/index.svg?label=docs)](https://huggingface.co/docs/safetensors/index)
[![Codecov](https://codecov.io/github/huggingface/safetensors/coverage.svg?branch=main)](https://codecov.io/gh/huggingface/safetensors)
[![Downloads](https://static.pepy.tech/badge/safetensors/month)](https://pepy.tech/project/safetensors)

Rust
[![Crates.io](https://img.shields.io/crates/v/safetensors.svg)](https://crates.io/crates/safetensors)
[![Documentation](https://docs.rs/safetensors/badge.svg)](https://docs.rs/safetensors/)
[![Codecov](https://codecov.io/github/huggingface/safetensors/coverage.svg?branch=main)](https://codecov.io/gh/huggingface/safetensors)
[![Dependency status](https://deps.rs/repo/github/huggingface/safetensors/status.svg?path=safetensors)](https://deps.rs/repo/github/huggingface/safetensors?path=safetensors)

# safetensors

## Safetensors

This repository implements a new simple format for storing tensors
safely (as opposed to pickle) and that is still fast (zero-copy).

### Installation
#### Pip

You can install safetensors via the pip manager:

```bash
pip install safetensors
```

#### From source

For the sources, you need Rust

```bash
# Install Rust
curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh
# Make sure it's up to date and using stable channel
rustup update
git clone https://github.com/huggingface/safetensors
cd safetensors/bindings/python
pip install setuptools_rust
pip install -e .
```

### Getting started

```python
import torch
from safetensors import safe_open
from safetensors.torch import save_file

tensors = {
   "weight1": torch.zeros((1024, 1024)),
   "weight2": torch.zeros((1024, 1024))
}
save_file(tensors, "model.safetensors")

tensors = {}
with safe_open("model.safetensors", framework="pt", device="cpu") as f:
   for key in f.keys():
       tensors[key] = f.get_tensor(key)
```

[Python documentation](https://huggingface.co/docs/safetensors/index)


### Format

- 8 bytes: `N`, an unsigned little-endian 64-bit integer, containing the size of the header
- N bytes: a JSON UTF-8 string representing the header.
  - The header data MUST begin with a `{` character (0x7B).
  - The header data MAY be trailing padded with whitespace (0x20).
  - The header is a dict like `{"TENSOR_NAME": {"dtype": "F16", "shape": [1, 16, 256], "data_offsets": [BEGIN, END]}, "NEXT_TENSOR_NAME": {...}, ...}`,
    - `data_offsets` point to the tensor data relative to the beginning of the byte buffer (i.e. not an absolute position in the file),
      with `BEGIN` as the starting offset and `END` as the one-past offset (so total tensor byte size = `END - BEGIN`).
  - A special key `__metadata__` is allowed to contain free form string-to-string map. Arbitrary JSON is not allowed, all values must be strings.
- Rest of the file: byte-buffer.

Notes:
 - Duplicate keys are disallowed. Not all parsers may respect this.
 - In general the subset of JSON is implicitly decided by `serde_json` for
   this library. Anything obscure might be modified at a later time, that odd ways
   to represent integer, newlines and escapes in utf-8 strings. This would only
   be done for safety concerns
 - Tensor values are not checked against, in particular NaN and +/-Inf could
   be in the file
 - Empty tensors (tensors with 1 dimension being 0) are allowed.
   They are not storing any data in the databuffer, yet retaining size in the header.
   They don't really bring a lot of values but are accepted since they are valid tensors
   from traditional tensor libraries perspective (torch, tensorflow, numpy, ..).
 - 0-rank Tensors (tensors with shape `[]`) are allowed, they are merely a scalar.
 - The byte buffer needs to be entirely indexed, and cannot contain holes. This prevents
   the creation of polyglot files.
 - Endianness: Little-endian.
   moment.
 - Order: 'C' or row-major.
 - Notes: Some smaller than 1 byte dtypes appeared, which make alignment tricky. Non traditional APIs might be required for those.


### Yet another format ?

The main rationale for this crate is to remove the need to use
`pickle` on `PyTorch` which is used by default.
There are other formats out there used by machine learning and more general
formats.


Let's take a look at alternatives and why this format is deemed interesting.
This is my very personal and probably biased view:

| Format                  | Safe | Zero-copy | Lazy loading | No file size limit | Layout control | Flexibility | Bfloat16/Fp8
| ----------------------- | --- | --- | --- | --- | --- | --- | --- |
| pickle (PyTorch)        | ✗ | ✗ | ✗ | 🗸 | ✗ | 🗸 | 🗸 |
| H5 (Tensorflow)         | 🗸 | ✗ | 🗸 | 🗸 | ~ | ~ | ✗ |
| SavedModel (Tensorflow) | 🗸 | ✗ | ✗ | 🗸 | 🗸 | ✗ | 🗸 |
| MsgPack (flax)          | 🗸 | 🗸 | ✗ | 🗸 | ✗ | ✗ | 🗸 |
| Protobuf (ONNX)         | 🗸 | ✗ | ✗ | ✗ | ✗ | ✗ | 🗸 |
| Cap'n'Proto             | 🗸 | 🗸 | ~ | 🗸 | 🗸 | ~ | ✗ |
| Arrow                   | ? | ? | ? | ? | ? | ? | ✗ |
| Numpy (npy,npz)         | 🗸 | ? | ? | ✗ | 🗸 | ✗ | ✗ |
| pdparams (Paddle)       | ✗ | ✗ | ✗ | 🗸 | ✗ | 🗸 | 🗸 |
| SafeTensors             | 🗸 | 🗸 | 🗸 | 🗸 | 🗸 | ✗ | 🗸 |

- Safe: Can I use a file randomly downloaded and expect not to run arbitrary code ?
- Zero-copy: Does reading the file require more memory than the original file ?
- Lazy loading: Can I inspect the file without loading everything ? And loading only
  some tensors in it without scanning the whole file (distributed setting) ?
- Layout control: Lazy loading, is not necessarily enough since if the information about tensors is spread out in your file, then even if the information is lazily accessible you might have to access most of your file to read the available tensors (incurring many DISK -> RAM copies). Controlling the layout to keep fast access to single tensors is important.
- No file size limit: Is there a limit to the file size ?
- Flexibility: Can I save custom code in the format and be able to use it later with zero extra code ? (~ means we can store more than pure tensors, but no custom code)
- Bfloat16/Fp8: Does the format support native bfloat16/fp8 (meaning no weird workarounds are
  necessary)? This is becoming increasingly important in the ML world.


### Main oppositions

- Pickle: Unsafe, runs arbitrary code
- H5: Apparently now discouraged for TF/Keras. Seems like a great fit otherwise actually. Some classic use after free issues: <https://www.cvedetails.com/vulnerability-list/vendor_id-15991/product_id-35054/Hdfgroup-Hdf5.html>. On a very different level than pickle security-wise. Also 210k lines of code vs ~400 lines for this lib currently.
- SavedModel: Tensorflow specific (it contains TF graph information).
- MsgPack: No layout control to enable lazy loading (important for loading specific parts in distributed setting)
- Protobuf: Hard 2Go max file size limit
- Cap'n'proto: Float16 support is not present [link](https://capnproto.org/language.html#built-in-types) so using a manual wrapper over a byte-buffer would be necessary. Layout control seems possible but not trivial as buffers have limitations [link](https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize).
- Numpy (npz): No `bfloat16` support. Vulnerable to zip bombs (DOS). Not zero-copy.
- Arrow: No `bfloat16` support.

### Notes

- Zero-copy: No format is really zero-copy in ML, it needs to go from disk to RAM/GPU RAM (that takes time). On CPU, if the file is already in cache, then it can
  truly be zero-copy, whereas on GPU there is not such disk cache, so a copy is always required
  but you can bypass allocating all the tensors on CPU at any given point.
  SafeTensors is not zero-copy for the header. The choice of JSON is pretty arbitrary, but since deserialization is <<< of the time required to load the actual tensor data and is readable I went that way, (also space is <<< to the tensor data).

- Endianness: Little-endian. This can be modified later, but it feels really unnecessary at the
  moment.
- Order: 'C' or row-major. This seems to have won. We can add that information later if needed.
- Stride: No striding, all tensors need to be packed before being serialized. I have yet to see a case where it seems useful to have a strided tensor stored in serialized format.
 - Sub 1 bytes dtypes: Dtypes can now have lower than 1 byte size, this makes alignment&adressing tricky. For now, the library will simply error out whenever an operation triggers an non aligned read. Trickier API may be created later for those non standard ops. 

### Benefits

Since we can invent a new format we can propose additional benefits:

- Prevent DOS attacks: We can craft the format in such a way that it's almost
  impossible to use malicious files to DOS attack a user. Currently, there's a limit
  on the size of the header of 100MB to prevent parsing extremely large JSON.
  Also when reading the file, there's a guarantee that addresses in the file
  do not overlap in any way, meaning when you're loading a file you should never
  exceed the size of the file in memory

- Faster load: PyTorch seems to be the fastest file to load out in the major
  ML formats. However, it does seem to have an extra copy on CPU, which we
  can bypass in this lib by using `torch.UntypedStorage.from_file`.
  Currently, CPU loading times are extremely fast with this lib compared to pickle.
  GPU loading times are as fast or faster than PyTorch equivalent.
  Loading first on CPU with memmapping with torch, and then moving all tensors to GPU seems
  to be faster too somehow (similar behavior in torch pickle)

- Lazy loading: in distributed (multi-node or multi-gpu) settings, it's nice to be able to
  load only part of the tensors on the various models. For
  [BLOOM](https://huggingface.co/bigscience/bloom) using this format enabled
  to load the model on 8 GPUs from 10mn with regular PyTorch weights down to 45s.
  This really speeds up feedbacks loops when developing on the model. For instance
  you don't have to have separate copies of the weights when changing the distribution
  strategy (for instance Pipeline Parallelism vs Tensor Parallelism).

License: Apache-2.0

The purpose of this directory is to showcase various attacks (and creating your own).


# Torch Arbitrary code execution

Try it out. This will create a seemingly innocuous `torch_ace.pt` file.
```
python torch_ace_create.py
python torch_ace_get_pwned.py
```

# PaddlePaddle Arbitrary code execution

Try it out. This will create a seemingly innocuous `paddle_ace.pdparams` file.
```
python paddle_ace_create.py
python paddle_ace_get_pwned.py
```

# Tensorflow (Keras) Arbitrary Code execution (does not affect `transformers`)

Try it out. This will create a seemingly innocuous `tf_ace.h5` file.
```
python tf_dos_create.py
python tf_dos_get_pwned.py
```

# Torch Denial of Service (OOM kills the running process)

Try it out. This will create a seemingly innocuous `torch_dos.pt` file.
```
python torch_dos_create.py
python torch_dos_get_pwned.py
```

# Numpy Denial of Service (OOM kills the running process)

Try it out. This will create a seemingly innocuous `numpy_dos.npz` file.
```
python numpy_dos_create.py
python numpy_dos_get_pwned.py
```

# Safetensors abuse attempts

In order to try and check the limits, we also try to abuse the current format.
Please send ideas!

A few things can be abused:
- Proposal 1: The initial 8 bytes, which could be too big with regards to the file. This crashes, and crashes early (Out of bounds) (Attempt #1).
- Proposal 2: The initial header is JSON, an attacker could use a 4Go JSON file to delay the loads. Debattable how much of an attack this is, but at least 
  it's impossible to "bomb" (like the DOS attacks above) where the files are vastly smaller than their expanded version (because of zip abuse).
  Various "protections" could be put in place, like a header proportion cap (header should always be <<< of the size of the file). (Attempt #2)
- Proposal 3: The offsets could be negative, out of the file. This is all crashing by default.
- Proposal 4: The offsets could overlap. ~~This is actually OK.~~ This is NOT ok.
                While testing Proposal 2, I realized that the tensors themselves where all allocated, and gave me an idea for a DOS exploit where you would have a relatively small
                file a few megs tops, but defining many tensors on the same overlapping part of the file, it was essentially a DOS attack. The mitigation is rather simple, we sanitize the fact
                that the offsets must be contiguous and non overlapping.
- Proposal 5: The offsets could mismatch the declared shapes + dtype. This validated against.
- Proposal 6: The file being mmaped could be modified while it's opened (attacker has access to your filesystem, seems like you're already pwned).
- Proposal 7: serde JSON deserialization abuse (nothing so far: https://cve.mitre.org/cgi-bin/cvekey.cgi?keyword=serde). It doesn't mean there isn't a flaw. Same goes for the actual rust compiled binary.

```
python safetensors_abuse_attempt_1.py
python safetensors_abuse_attempt_2.py
python safetensors_abuse_attempt_3.py
```

## Installation

```
pip install safetensors
```


## Usage

### Numpy

```python
from safetensors.numpy import save_file, load_file
import numpy as np

tensors = {
   "a": np.zeros((2, 2)),
   "b": np.zeros((2, 3), dtype=np.uint8)
}

save_file(tensors, "./model.safetensors")


# Now loading
loaded = load_file("./model.safetensors")
```

### Torch

```python
from safetensors.torch import save_file, load_file
import torch

tensors = {
   "a": torch.zeros((2, 2)),
   "b": torch.zeros((2, 3), dtype=torch.uint8)
}

save_file(tensors, "./model.safetensors")


# Now loading
loaded = load_file("./model.safetensors")
```

### Developing

```
# inside ./safetensors/bindings/python
pip install .[dev]
```
Should be enough to install this library locally.

### Testing

```
# inside ./safetensors/bindings/python
pip install .[dev]
pytest -sv tests/
```

#' Writes a list of tensors to the safetensors format
#'
#' @param tensors A named list of tensors. Currently only torch tensors are supported.
#' @param path The path to save the tensors to. It can also be a binary connection, as eg,
#'   created with `file()`.
#' @param ... Currently unused.
#' @param metadata An optional string that is added to the file header. Possibly
#'   adding additional description to the weights.
#'
#' @examples
#' if (rlang::is_installed("torch") && torch::torch_is_installed()) {
#'   tensors <- list(x = torch::torch_randn(10, 10))
#'   temp <- tempfile()
#'   safe_save_file(tensors, temp)
#'   safe_load_file(temp, framework = "torch")
#'
#'   ser <- safe_serialize(tensors)
#' }
#'
#' @returns The path invisibly or a raw vector.
#'
#' @export
safe_save_file <- function(tensors, path, ..., metadata = NULL) {
  if (any(duplicated(names(tensors)))) {
    cli::cli_abort("Duplicated names are not allowed in {.arg tensors}")
  }

  if (is.character(path)) {
    con <- file(path, open = "wb")
    on.exit(
      {
        close(con)
      },
      add = TRUE
    )
  } else {
    con <- path
  }

  write_safe(tensors, metadata, con)
  invisible(path)
}

#' @describeIn safe_save_file Serializes the tensors and returns a raw vector.
#' @export
safe_serialize <- function(tensors, ..., metadata = NULL) {
  con <- rawConnection(raw(), open = "wb")
  on.exit(
    {
      close(con)
    },
    add = TRUE
  )
  safe_save_file(tensors, con, metadata = metadata)
  rawConnectionValue(con)
}

write_safe <- function(tensors, metadata, con) {
  meta <- make_meta(tensors, metadata)
  meta_raw <- charToRaw(jsonlite::toJSON(meta, auto_unbox = TRUE))
  # write the metadatasize as a 64bit int

  writeBin(length(meta_raw), con = con, size = 8L)
  writeBin(meta_raw, con = con)
  for (tensor in tensors) {
    buf <- safe_tensor_buffer(tensor)
    writeBin(buf, con = con)
  }
}

make_meta <- function(tensors, metadata) {
  meta_ <- structure(
    vector(mode = "list", length = length(tensors)),
    names = names(tensors)
  )

  if (!is.null(metadata)) {
    meta_[["__metadata__"]] <- validate_metadata(metadata)
  }

  pos <- 0L
  for (nm in names(tensors)) {
    meta <- safe_tensor_meta(tensors[[nm]])
    meta$data_offsets <- c(pos, pos + size_from_meta(meta))
    pos <- meta$data_offsets[2]
    meta_[[nm]] <- meta
  }

  meta_
}

#' @title Get raw buffer from a tensor
#' @description
#' Convert a tensor object to a raw buffer in the formated expected by safetensors.
#' @param x (any)\cr
#'   Tensor object.
#' @returns (`raw`)
#' @export
safe_tensor_buffer <- function(x) {
  UseMethod("safe_tensor_buffer")
}

#' @title Get metadata from a tensor
#' @description
#' Get the metadata from a tensor.
#' @param x (any)\cr
#'   Tensor object.
#' @returns (`list`)
#' @export
safe_tensor_meta <- function(x) {
  UseMethod("safe_tensor_meta")
}

size_from_meta <- function(meta) {
  numel <- prod(as.numeric(meta$shape))

  el_size <- if (meta$dtype == "F32") {
    4L
  } else if (meta$dtype == "F16") {
    2L
  } else if (meta$dtype == "F64") {
    8L
  } else if (meta$dtype == "U8") {
    1L
  } else if (meta$dtype == "I8") {
    1L
  } else if (meta$dtype == "I16") {
    2L
  } else if (meta$dtype == "I32") {
    4L
  } else if (meta$dtype == "I64") {
    8L
  } else if (meta$dtype == "BOOL") {
    1L
  } else if (meta$dtype == "U16") {
    2L
  } else if (meta$dtype == "U32") {
    4L
  } else if (meta$dtype == "U64") {
    8L
  } else {
    cli::cli_abort("Unsupported dtype {.val {meta$dtype}}")
  }

  as.integer(numel * el_size)
}

validate_metadata <- function(x) {
  if (!rlang::is_list(x)) {
    cli::cli_abort("{.arg metadata} must be a list.")
  }
  if (!rlang::is_named(x)) {
    cli::cli_abort("{.arg metadata} must be a named list.")
  }
  lapply(x, function(item) {
    if (!rlang::is_scalar_character(item)) {
      cli::cli_abort(
        "{.arg metadata} must be a named list of scalar characters."
      )
    }
  })
  x
}

test_that("can write a safetensors file (torch)", {
  skip_if_not_installed("torch")

  tensors <- list(
    x = torch::torch_randn(10, 10),
    y = torch::torch_randn(5, 5)
  )

  tmp <- tempfile(fileext = ".safetensors")
  safe_save_file(tensors, tmp)

  reloaded <- safe_load_file(tmp, framework = "torch")

  expect_true(torch::torch_allclose(tensors$x, reloaded$x))
  expect_true(torch::torch_allclose(tensors$y, reloaded$y))
})

test_that("with different datatypes (torch)", {
  data_type <- c(
    "float16",
    "float",
    "float64",
    "bool",
    "uint8",
    "int8",
    "int16",
    "int32",
    "int64"
  )

  for (dtype in data_type) {
    x <- list(x = torch::torch_randn(10)$to(dtype = dtype))

    tmp <- tempfile(fileext = ".safetensors")
    safe_save_file(x, tmp)

    reloaded <- safe_load_file(tmp, framework = "torch")

    expect_true(torch::torch_allclose(x$x, reloaded$x))
  }
})

test_that("metadata validations", {
  tensors <- list(
    x = torch::torch_randn(10, 10),
    y = torch::torch_randn(5, 5)
  )

  tmp <- tempfile(fileext = ".safetensors")

  metadata <- 1
  expect_snapshot_error({
    safe_save_file(tensors, tmp, metadata = metadata)
  })

  metadata <- list(x = 1)
  expect_snapshot_error({
    safe_save_file(tensors, tmp, metadata = metadata)
  })

  metadata <- list(y = c("1", "2"))
  expect_snapshot_error({
    safe_save_file(tensors, tmp, metadata = metadata)
  })

  metadata <- list("a")
  expect_snapshot_error({
    safe_save_file(tensors, tmp, metadata = metadata)
  })
})

test_that("load a file (torch)", {
  skip_if_not_installed("torch")
  dict <- safe_load_file(
    test_path("safetensors/hello.safetensors"),
    framework = "torch"
  )
  expect_equal(names(dict), c("hello", "world"))

  expect_equal(dict$hello$shape, c(10, 10))
  expect_true(all(as.numeric(dict$hello) == 1))

  expect_equal(dict$world$shape, c(5, 10))
  expect_true(all(as.numeric(dict$world) == 0))
})

#test_that("torch & pjrt interoperability", {
#  skip_if_not_installed("torch")
#  skip_if_not_installed("pjrt")
#  skip_on_os("windows")
#
#  x <- matrix(c(1, 2, 3, 4, 5, 6), nrow = 2)
#  torch_tensors <- list(x = torch::torch_tensor(x))
#  pjrt_buffers <- list(x = pjrt::pjrt_buffer(x))
#
#  temp1 <- tempfile()
#  temp2 <- tempfile()
#
#  safe_save_file(torch_tensors, temp1)
#  safe_save_file(pjrt_buffers, temp2)
#
#  x1 <- safe_load_file(temp1, framework = "torch")
#  x2 <- safe_load_file(temp2, framework = "pjrt")
#  x3 <- safe_load_file(temp1, framework = "pjrt")
#  x4 <- safe_load_file(temp2, framework = "torch")
#
#  expect_equal(torch::as_array(x1$x), pjrt::as_array(x2$x))
#  expect_equal(pjrt::as_array(x2$x), pjrt::as_array(x3$x))
#  expect_equal(pjrt::as_array(x3$x), torch::as_array(x4$x))
#})

torch_dtype_to_safe <- function(x) {
  if (x == torch::torch_float()) {
    return("F32")
  } else if (x == torch::torch_float16()) {
    return("F16")
  } else if (x == torch::torch_float64()) {
    return("F64")
  } else if (x == torch::torch_bool()) {
    return("BOOL")
  } else if (x == torch::torch_uint8()) {
    return("U8")
  } else if (x == torch::torch_int8()) {
    return("I8")
  } else if (x == torch::torch_int16()) {
    return("I16")
  } else if (x == torch::torch_int32()) {
    return("I32")
  } else if (x == torch::torch_int64()) {
    return("I64")
  } else if (x == torch::torch_bfloat16()) {
    return("BF16")
  } else if (x == torch::torch_cfloat()) {
    return("C64")
  } else if (x == torch::torch_cdouble()) {
    return("C128")
  } else {
    cli::cli_abort("Unsupported data type {.val {x}}")
  }
}

#' @export
safe_tensor_buffer.torch_tensor <- function(x) {
  torch::buffer_from_torch_tensor(x$cpu())
}

#' @export
safe_tensor_meta.torch_tensor <- function(x) {
  list(
    shape = as.list(x$shape), # we must store as a list to avoid simplification
    dtype = torch_dtype_to_safe(x$dtype)
  )
}

torch_tensor_from_raw <- function(raw, meta, device = "cpu") {
  x <- torch::torch_tensor_from_buffer(
    raw,
    shape = meta$shape,
    dtype = torch_dtype_from_safe(meta$dtype)
  )
  if (device == "cpu") {
    x$clone() # we need to explicitly clone in case the device is cpu
  } else {
    x$to(device = device)
  }
}

torch_dtype_from_safe <- function(x) {
  switch(
    x,
    "F16" = "float16",
    "F32" = "float",
    "F64" = "float64",
    "BOOL" = "bool",
    "U8" = "uint8",
    "I8" = "int8",
    "I16" = "int16",
    "I32" = "int32",
    "I64" = "int64",
    "BF16" = "bfloat16",
    cli::cli_abort("Unsupported dtype {.val {x}}")
  )
}

#' @title Reflection of supported frameworks
#' @description
#' A reflection of supported frameworks.
#' @export
safetensors_frameworks <- new.env()

safetensors_frameworks[["torch"]] <- list(
  constructor = torch_tensor_from_raw,
  packages = "torch"
)

#' Safe load a safetensors file
#'
#' Loads an safetensors file from disk.
#'
#' @param path Path to the file to load
#' @param framework Framework to load the data into. Currently supports "torch" and "pjrt"
#' @param ... Additional framework dependent arguments to pass to the tensor creation function.
#'
#' @returns A list with tensors in the file. The `metadata` attribute can be used
#' to find metadata the metadata header in the file.
#'
#' @examples
#' if (rlang::is_installed("torch") && torch::torch_is_installed()) {
#'   tensors <- list(x = torch::torch_randn(10, 10))
#'   temp <- tempfile()
#'   safe_save_file(tensors, temp)
#'   safe_load_file(temp, framework = "torch")
#' }
#'
#' @seealso [safetensors], [safe_save_file()]
#'
#' @export
safe_load_file <- function(path, ..., framework) {
  f <- safetensors$new(path, ..., framework = framework)
  nms <- f$keys()
  output <- structure(
    vector(length = length(nms), mode = "list"),
    names = nms,
    metadata = f$metadata
  )
  for (key in nms) {
    output[[key]] <- f$get_tensor(key)
  }
  attr(output, "max_offset") <- f$max_offset
  output
}

#' Low level control over safetensors files
#'
#' Allows opening a connection to a safetensors file and query the tensor names,
#' metadata, etc.
#' Opening a connection only reads the file metadata into memory.
#' This allows for more fined grained control over reading.
#'
#' @examples
#' if (rlang::is_installed("torch") && torch::torch_is_installed()) {
#' tensors <- list(x = torch::torch_randn(10, 10))
#' temp <- tempfile()
#' safe_save_file(tensors, temp)
#' f <- safetensors$new(temp, framework = "torch")
#' f$get_tensor("x")
#' }
#'
#' @importFrom R6 R6Class
#'
#' @export
safetensors <- R6::R6Class(
  "safetensors",
  public = list(
    #' @field con the connection object with the file
    con = NULL,
    #' @field metadata an R list containing the metadata header in the file
    metadata = NULL,
    #' @field framework the framework used to return the tensors
    framework = NULL,
    #' @field args additional arguments for tensor creation
    args = NULL,
    #' @field max_offset the largest offset boundary that was visited. Mainly
    #' used in torch to find the end of the safetensors file.
    max_offset = 0L,
    #' @description
    #' Opens the connection with the file
    #' @param path Path to the file to load
    #' @param framework Framework to load the data into. Currently supports "torch" and "pjrt"
    #' @param ... (any)\cr
    #'   Additional, framework dependent, arguments to pass to use when creating the tensor.
    #'   For torch, this is the device, for pjrt the client.
    initialize = function(path, ..., framework) {
      self$framework <- validate_framework(framework)
      self$args <- list(...)

      # read in the metadata and store it
      if (is.raw(path)) {
        self$con <- rawConnection(path, open = "rb")
      } else if (is.character(path)) {
        self$con <- file(path, "rb")
      } else if (inherits(path, "connection")) {
        # safetensors has no responsability over this connection as this was
        # created efore passing to it.
        private$close_con <- FALSE
        self$con <- path
      }

      metadata_size <- readBin(self$con, what = integer(), n = 1, size = 8)
      raw_json <- readBin(self$con, what = "raw", n = metadata_size)

      self$metadata <- jsonlite::parse_json(
        rawToChar(raw_json),
        simplifyVector = TRUE
      )
      private$byte_buffer_begin <- 8L + metadata_size
    },
    #' @description
    #' Get the keys (tensor names) in the file
    keys = function() {
      keys <- names(self$metadata)
      keys[keys != "__metadata__"]
    },
    #' @description
    #' Get a tensor from its name
    #' @param name Name of the tensor to load
    get_tensor = function(name) {
      meta <- self$metadata[[name]]

      offset_start <- private$byte_buffer_begin + meta$data_offsets[1]
      offset_length <- meta$data_offsets[2] - meta$data_offsets[1]
      self$max_offset <- max(self$max_offset, offset_start + offset_length)

      seek(self$con, offset_start)
      raw_tensor <- readBin(self$con, what = "raw", n = offset_length)

      if (!self$framework %in% names(safetensors_frameworks)) {
        cli::cli_abort("Unsupported framework {.val {.self$framework}}")
      }

      rlang::exec(
        safetensors_frameworks[[self$framework]]$constructor,
        raw_tensor,
        meta,
        !!!self$args
      )
    }
  ),
  private = list(
    byte_buffer_begin = 0L,
    close_con = TRUE,
    finalize = function() {
      if (private$close_con) {
        close(self$con)
      }
    }
  )
)

validate_framework <- function(x) {
  info <- safetensors_frameworks[[x]]

  if (is.null(info)) {
    cli::cli_abort("Unsupported framework {.val {x}}")
  }
  rlang::check_installed(info$packages, reason = "for loading {x} tensors.")
  x
}

set -ex



pip check
cd bindings/python
pytest -v tests --ignore=tests/test_tf_comparison.py --ignore=tests/test_paddle_comparison.py --ignore=tests/test_flax_comparison.py --ignore=tests/test_mlx_comparison.py
exit 0

pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
cd bindings/python
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -v tests --ignore=tests/test_tf_comparison.py --ignore=tests/test_paddle_comparison.py --ignore=tests/test_flax_comparison.py --ignore=tests/test_mlx_comparison.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0

# What does this PR do?

<!--
Congratulations! You've made it this far! You're not quite done yet though.

Once merged, your PR is going to appear in the release notes with the title you set, so make sure it's a great title that fully reflects the extent of your awesome contribution.

Then, please replace this with a description of the change and which issue is fixed (if applicable). Please also include relevant motivation and context. List any dependencies (if any) that are required for this change.

Once you're done, someone will review your PR shortly (see the section "Who can review?" below to tag some potential reviewers). They may suggest changes to make the code even better. If no one reviewed your PR after a week has passed, don't hesitate to post a new comment @-mentioning the same persons---sometimes notifications get lost.
-->

<!-- Remove if not applicable -->

Fixes # (issue) or description of the problem this PR solves.

# What does this PR do?

<!--
Congratulations! You've made it this far! You're not quite done yet though.

Once merged, your PR is going to appear in the release notes with the title you set, so make sure it's a great title that fully reflects the extent of your awesome contribution.

Then, please replace this with a description of the change and which issue is fixed (if applicable). Please also include relevant motivation and context. List any dependencies (if any) that are required for this change.

Once you're done, someone will review your PR shortly (see the section "Who can review?" below to tag some potential reviewers). They may suggest changes to make the code even better. If no one reviewed your PR after a week has passed, don't hesitate to post a new comment @-mentioning the same persons---sometimes notifications get lost.
-->

<!-- Remove if not applicable -->

Fixes # (issue) or description of the problem this PR solves.

## AI model use

We do accept PRs that were developed together with an AI model, but we ask
that you to fill out this section.

- [ ] No AI model was used when making this PR.
- [ ] An AI model assisted me in developing this PR.
- [ ] Development of this PR was done by an AI model.

If you ticked one of the last two options, please state the model and model
version that you used:

If you ticked the last option, please list the prompt(s) that you used:

# Convert weights to safetensors

PyTorch model weights are commonly saved and stored as `.bin` files with Python's [`pickle`](https://docs.python.org/3/library/pickle.html) utility. To save and store your model weights in the more secure `safetensor` format, we recommend converting your weights to `.safetensors`.

The easiest way to convert your model weights is to use the [Convert Space](https://huggingface.co/spaces/safetensors/convert), given your model weights are already stored on the Hub. The Convert Space downloads the pickled weights, converts them, and opens a Pull Request to upload the newly converted `.safetensors` file to your repository.

<Tip warning={true}>

For larger models, the Space may be a bit slower because its resources are tied up in converting other models. You can also try running the [convert.py](https://github.com/huggingface/safetensors/blob/main/bindings/python/convert.py) script (this is what the Space is running) locally to convert your weights.

Feel free to ping [@Narsil](https://huggingface.co/Narsil) for any issues with the Space.

</Tip>

# Metadata Parsing

Given the simplicity of the format, it's very simple and efficient to fetch and parse metadata about Safetensors weights – i.e. the list of tensors, their types, and their shapes or numbers of parameters – using small [(Range) HTTP requests](https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests).

This parsing has been implemented in JS in [`huggingface.js`](https://huggingface.co/docs/huggingface.js/main/en/hub/modules#parsesafetensorsmetadata) (sample code follows below), but it would be similar in any language.

## Example use case

There can be many potential use cases. For instance, we use it on the HuggingFace Hub to display info about models which have safetensors weights:

<div class="flex justify-center">
    <img class="block dark:hidden" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/model-page-light.png"/>
    <img class="hidden dark:block" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/model-page-dark.png"/>
</div>

<div class="flex justify-center">
    <img class="block dark:hidden" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/view-all-tensors-light.png"/>
    <img class="hidden dark:block" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/view-all-tensors-dark.png"/>
</div>

## Usage

<hfoptions id="metadata">

<hfoption id="http">
From [🤗 Hub](hf.co/models), you can get metadata of a model with [HTTP range requests](https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests) instead of downloading the entire safetensors file with all the weights. In this example python script below (you can use any language that has HTTP requests support), we are parsing metadata of [gpt2](https://huggingface.co/gpt2/blob/main/model.safetensors). 

```python
import requests # pip install requests
import struct

def parse_single_file(url):
    # Fetch the first 8 bytes of the file
    headers = {'Range': 'bytes=0-7'}
    response = requests.get(url, headers=headers)
    # Interpret the bytes as a little-endian unsigned 64-bit integer
    length_of_header = struct.unpack('<Q', response.content)[0]
    # Fetch length_of_header bytes starting from the 9th byte
    headers = {'Range': f'bytes=8-{7 + length_of_header}'}
    response = requests.get(url, headers=headers)
    # Interpret the response as a JSON object
    header = response.json()
    return header

url = "https://huggingface.co/gpt2/resolve/main/model.safetensors"
header = parse_single_file(url)

print(header)
# {
#   "__metadata__": { "format": "pt" },
#   "h.10.ln_1.weight": {
#     "dtype": "F32",
#     "shape": [768],
#     "data_offsets": [223154176, 223157248]
#   },
#   ...
# }
```
</hfoption>

<hfoption id="javascript">
Using [`huggingface.js`](https://huggingface.co/docs/huggingface.js)

```ts
import { parseSafetensorsMetadata } from "@huggingface/hub";

const info = await parseSafetensorsMetadata({
	repo: { type: "model", name: "bigscience/bloom" },
});

console.log(info)
// {
//   sharded: true,
//   index: {
//     metadata: { total_size: 352494542848 },
//     weight_map: {
//       'h.0.input_layernorm.bias': 'model_00002-of-00072.safetensors',
//       ...
//     }
//   },
//   headers: {
//     __metadata__: {'format': 'pt'},
//     'h.2.attn.c_attn.weight': {'dtype': 'F32', 'shape': [768, 2304], 'data_offsets': [541012992, 548090880]},
//     ...
//   }
// }
```

Depending on whether the safetensors weights are sharded into multiple files or not, the output of the call above will be:

```ts
export type SafetensorsParseFromRepo =
| {
		sharded: false;
		header: SafetensorsFileHeader;
	}
| {
		sharded: true;
		index: SafetensorsIndexJson;
		headers: SafetensorsShardedHeaders;
	};
```

where the underlying `types` are the following:

```ts
type FileName = string;

type TensorName = string;
type Dtype = "F64" | "F32" | "F16" | "BF16" | "I64" | "I32" | "I16" | "I8" | "U8" | "BOOL";

interface TensorInfo {
	dtype: Dtype;
	shape: number[];
	data_offsets: [number, number];
}

type SafetensorsFileHeader = Record<TensorName, TensorInfo> & {
	__metadata__: Record<string, string>;
};

interface SafetensorsIndexJson {
	weight_map: Record<TensorName, FileName>;
}

export type SafetensorsShardedHeaders = Record<FileName, SafetensorsFileHeader>;
```
</hfoption>

<hfoption id="python">
[`huggingface_hub`](https://huggingface.co/docs/huggingface_hub) provides a Python API to parse safetensors metadata.
Use [`get_safetensors_metadata`](https://huggingface.co/docs/huggingface_hub/package_reference/hf_api#huggingface_hub.HfApi.get_safetensors_metadata) to get all safetensors metadata of a model.
Depending on if the model is sharded or not, one or multiple safetensors files will be parsed.

```python
>>> from huggingface_hub import get_safetensors_metadata

# Parse repo with single weights file
>>> metadata = get_safetensors_metadata("bigscience/bloomz-560m")
>>> metadata
SafetensorsRepoMetadata(
    metadata=None,
    sharded=False,
    weight_map={'h.0.input_layernorm.bias': 'model.safetensors', ...},
    files_metadata={'model.safetensors': SafetensorsFileMetadata(...)}
)
>>> metadata.files_metadata["model.safetensors"].metadata
{'format': 'pt'}

# Parse repo with sharded model (i.e. multiple weights files)
>>> metadata = get_safetensors_metadata("bigscience/bloom")
Parse safetensors files: 100%|██████████████████████████████████████████| 72/72 [00:12<00:00,  5.78it/s]
>>> metadata
SafetensorsRepoMetadata(metadata={'total_size': 352494542848}, sharded=True, weight_map={...}, files_metadata={...})
>>> len(metadata.files_metadata)
72  # All safetensors files have been fetched

# Parse repo that is not a safetensors repo
>>> get_safetensors_metadata("runwayml/stable-diffusion-v1-5")
NotASafetensorsRepoError: 'runwayml/stable-diffusion-v1-5' is not a safetensors repo. Couldn't find 'model.safetensors.index.json' or 'model.safetensors' files.
```

To parse the metadata of a single safetensors file, use [`parse_safetensors_file_metadata`](https://huggingface.co/docs/huggingface_hub/package_reference/hf_api#huggingface_hub.HfApi.parse_safetensors_file_metadata).
</hfoption>

</hfoptions>

## Example output

For instance, here are the number of params per dtype for a few models on the HuggingFace Hub. Also see [this issue](https://github.com/huggingface/safetensors/issues/44) for more examples of usage.

model | safetensors | params
--- | --- | ---
[gpt2](https://huggingface.co/gpt2?show_tensors=true) | single-file | { 'F32' => 137022720 }
[roberta-base](https://huggingface.co/roberta-base?show_tensors=true) | single-file | { 'F32' => 124697433, 'I64' => 514 }
[Jean-Baptiste/camembert-ner](https://huggingface.co/Jean-Baptiste/camembert-ner?show_tensors=true) | single-file | { 'F32' => 110035205, 'I64' => 514 }
[roberta-large](https://huggingface.co/roberta-large?show_tensors=true) | single-file | { 'F32' => 355412057, 'I64' => 514 }
[distilbert-base-german-cased](https://huggingface.co/distilbert-base-german-cased?show_tensors=true) | single-file | { 'F32' => 67431550 }
[EleutherAI/gpt-neox-20b](https://huggingface.co/EleutherAI/gpt-neox-20b?show_tensors=true) | sharded | { 'F16' => 20554568208, 'U8' => 184549376 }
[bigscience/bloom-560m](https://huggingface.co/bigscience/bloom-560m?show_tensors=true) | single-file | { 'F16' => 559214592 }
[bigscience/bloom](https://huggingface.co/bigscience/bloom?show_tensors=true) | sharded | { 'BF16' => 176247271424 }
[bigscience/bloom-3b](https://huggingface.co/bigscience/bloom-3b?show_tensors=true) | single-file | { 'F16' => 3002557440 }

# Torch API

[[autodoc]] safetensors.torch.load_file
[[autodoc]] safetensors.torch.load
[[autodoc]] safetensors.torch.save_file
[[autodoc]] safetensors.torch.save
[[autodoc]] safetensors.torch.load_model
[[autodoc]] safetensors.torch.save_model

# How to Contribute to Safetensors

Everyone is welcome to contribute, and we value everybody's contribution. Code contributions are not the only way to help the community: answering questions, helping others, and improving the documentation are also immensely valuable.

However you choose to contribute, please be mindful and respect our [code of conduct](CODE_OF_CONDUCT.md).

## Table of Contents

- [Ways to Contribute](#ways-to-contribute)
- [Fixing Bugs](#fixing-bugs)
- [Submitting a Feature Request](#submitting-a-feature-request)
- [Implementing a Feature](#implementing-a-feature)
- [Asking for Help](#asking-for-help)
- [Setting Up Your Development Environment](#setting-up-your-development-environment)
  - [Prerequisites](#prerequisites)
  - [Rust Core](#rust-core)
  - [Python Bindings](#python-bindings)
  - [Running Tests](#running-tests)
- [Style Guide](#style-guide)
  - [Rust](#rust)
  - [Python](#python)
- [Submitting a Pull Request](#submitting-a-pull-request)
- [Review Process](#review-process)

---

## Ways to Contribute

There are many ways to contribute to Safetensors:

- **Fix outstanding issues** in the [issue tracker](https://github.com/huggingface/safetensors/issues), especially those tagged [`good first issue`](https://github.com/huggingface/safetensors/issues?q=is%3Aopen+is%3Aissue+label%3A%22good+first+issue%22) or [`help wanted`](https://github.com/huggingface/safetensors/issues?q=is%3Aopen+is%3Aissue+label%3A%22help+wanted%22).
- **Submit bug reports or feature requests** by [opening a new issue](https://github.com/huggingface/safetensors/issues/new/choose).
- **Improve documentation** — if anything is unclear, incomplete, or missing, a PR is always appreciated.
- **Write tests** — more test coverage means more confidence in correctness.
- **Help others** by answering questions in [issues](https://github.com/huggingface/safetensors/issues).

---

## Fixing Bugs

If you find a bug and want to fix it yourself, please do! Start by opening an issue to describe the bug and let the maintainers know you're working on it — this prevents duplicate effort.

> **Do you have a security concern?** Please do **not** open a public issue. Instead, follow our [security policy](SECURITY.md) to report it responsibly.

When submitting a bug fix:

1. Link the relevant issue in your pull request description.
2. Include a test that would have caught the bug.
3. Keep the fix focused — avoid mixing unrelated changes in the same PR.

---

## Submitting a Feature Request

We love ideas! Before opening a feature request issue, please **search existing issues** to see if it's already been discussed.

A good feature request explains:
- The **problem** you're trying to solve (not just the proposed solution).
- Why you believe this belongs in the core library rather than userspace.
- Whether you're willing to implement it yourself.

Because safetensors deliberately aims to remain a **simple, minimal format**, we are conservative about adding new features. Features that add significant complexity to the spec or break backward compatibility will need a strong justification.

---

## Implementing a Feature

> We strongly recommend opening an issue or discussion before investing significant time implementing a new feature, especially if it touches the binary format specification. The maintainers can give early feedback and confirm the feature is likely to be accepted.

For **small, clearly-scoped improvements** (e.g., an ergonomic API addition in the Python bindings), you're welcome to open a PR directly.

For **spec-level changes** (new dtypes, changes to the header format, etc.), please open an issue first — these changes require careful design review and community input.

---

## Asking for Help

Don't hesitate to ask questions! You can:

- Open a [GitHub Discussion](https://github.com/huggingface/safetensors/discussions)
- Tag your StackOverflow question with `safetensors`

---

## Setting Up Your Development Environment

### Prerequisites

The safetensors repository contains a Rust core library and Python bindings (with optional support for PyTorch, NumPy, TensorFlow, JAX, and PaddlePaddle). You'll need:

- **Rust** (stable toolchain) — [install via rustup](https://rustup.rs/)
- **Python 3.10+**
- `pip` and optionally a virtual environment manager
- [optional] we strongly recommend installing uv for python management, but that's up to you!

```bash
# Install Rust
curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh
rustup update

# Clone the repo
git clone https://github.com/huggingface/safetensors.git
cd safetensors
```

### Rust Core

The core library lives in `safetensors/`. To build and test it:

```bash
cd safetensors

# Build
cargo build

# Run tests
cargo test

# Run tests with all features
cargo test --all-features

# Check formatting and lints
cargo fmt --check
cargo clippy --all-targets --all-features -- -D warnings
```

### Python Bindings

The Python bindings live in `bindings/python/`. They use [PyO3](https://pyo3.rs/) via `maturin`.

```bash
cd bindings/python

# Create and activate a virtual environment (recommended)
python -m venv .venv
source .venv/bin/activate  # On Windows: .venv\Scripts\activate

# Install development dependencies
pip install setuptools-rust maturin
pip install -e ".[dev]"

# Build the Rust extension in-place (needed after changing Rust code)
pip install -e .
# or, for faster incremental builds during development:
maturin develop
```

> **Tip:** After every change to the Rust code, re-run `maturin develop` (or `pip install -e .`) before running Python tests — otherwise you'll be testing stale compiled code. You can also add the `--release` flag when building with maturin for a proper "release" build, which will be faster than the dev build.

#### Installing ML Framework Dependencies

Install the frameworks you want to test against:

```bash
# PyTorch
pip install torch --index-url https://download.pytorch.org/whl/cpu

# NumPy
pip install numpy

# TensorFlow
pip install tensorflow-cpu

# JAX
pip install jax

# PaddlePaddle
pip install paddlepaddle
```

### Running Tests

#### Rust Tests

```bash
# From the repo root
cargo test                        # core library tests
cargo test --manifest-path safetensors/Cargo.toml --all-features
```

#### Python Tests

```bash
cd bindings/python

# Run all Python tests
pytest tests/

# Run a specific test file
pytest tests/test_torch_serialization.py

# Run with verbose output
pytest -v tests/

# Run and stop at first failure
pytest -x tests/
```

#### Testing Across Frameworks

If you've installed multiple frameworks, run the full test suite to make sure your change doesn't break any of them:

```bash
pytest tests/ -v --tb=short
```

---

## Style Guide

Please match the existing code style. We use automated formatters and linters — a PR that fails CI checks will need to be fixed before it can be reviewed.

### Rust

We follow standard Rust conventions as enforced by the Rust toolchain:

```bash
# Format code
cargo fmt

# Lint
cargo clippy --all-targets --all-features -- -D warnings
```

- Use `cargo fmt` before every commit.
- Fix all `clippy` warnings — warnings are treated as errors in CI.
- Write documentation comments (`///`) for all public types and functions.
- Add unit tests inline in the same file (`#[cfg(test)]`).

### Python

We use [`ruff`](https://github.com/astral-sh/ruff) for both linting and formatting:

```bash
pip install ruff

# Check linting
ruff check bindings/python/

# Format code
ruff format bindings/python/

# Check formatting without modifying files
ruff format --check bindings/python/
```

Additional guidelines:

- Public Python APIs should have docstrings in [NumPy docstring style](https://numpydoc.readthedocs.io/en/latest/format.html).
- Type annotations are encouraged for all new code.
- Keep the Python surface area small and consistent — prefer matching the existing API style over introducing new patterns.

---

## Submitting a Pull Request

When you're ready to submit your contribution:

1. **Fork** the repository and create a branch from `main`:
   ```bash
   git checkout -b your-username/fix-brief-description
   ```

2. **Make your changes**, following the style guide above.

3. **Add or update tests** for any code you change. PRs that reduce test coverage are unlikely to be accepted.

4. **Update documentation** if your change affects user-facing behavior (docstrings, `README.md`, `docs/`).

5. **Run the full test suite** locally and make sure everything passes before pushing.

6. **Open a pull request** against the `main` branch. In the description:
   - Summarize **what** you changed and **why**.
   - Link any related issues (e.g., `Closes #123` or `Fixes #456`).
   - Mention any areas of the code you're uncertain about or that you'd like reviewers to pay special attention to.
   - We accept PRs that are made with the help of an AI model, but would appreciate if you
     fill out the AI model-related fields in the description.

7. **Sign your commits** if your organization requires it (optional otherwise). All commits merged into `main` should have a clean history.

> **Work in progress?** Open a [Draft PR](https://github.blog/2019-02-14-introducing-draft-pull-requests/) to share early progress and get feedback before the implementation is complete.

---

## Review Process

We aim to review PRs within a few business days. The review process typically works like this:

1. A maintainer will be assigned to review your PR.
2. They may request changes or ask clarifying questions — please respond to comments as promptly as you can.
3. Once all feedback is addressed and CI passes, a maintainer will approve and merge the PR.

**A few things that speed up the review:**

- Keep PRs small and focused on a single concern. Large, sprawling PRs take much longer to review.
- Write a clear, descriptive PR description — it helps reviewers understand intent and context without having to ask.
- Don't resolve review comments yourself (leave that to the reviewer) so the conversation thread stays readable.
- Be patient and respectful. Maintainers are often juggling many things at once.

We appreciate every contribution, big or small. Thank you for taking the time to make Safetensors better! 🤗

# Speed Comparison

<a href="https://colab.research.google.com/github/huggingface/notebooks/blob/main/safetensors_doc/en/speed.ipynb" target="_blank" class="absolute z-10 right-0 top-0">
    <img
        alt="Open In Colab"
        class="!m-0"
        src="https://colab.research.google.com/assets/colab-badge.svg"
    />
</a>

`Safetensors` is really fast. Let's compare it against `PyTorch` by loading [gpt2](https://huggingface.co/gpt2) weights. To run the [GPU benchmark](#gpu-benchmark), make sure your machine has GPU or you have selected `GPU runtime` if you are using Google Colab.

Before you begin, make sure you have all the necessary libraries installed:

```bash
pip install safetensors huggingface_hub torch
```

Let's start by importing all the packages that will be used:

```py
>>> import os
>>> import datetime
>>> from huggingface_hub import hf_hub_download
>>> from safetensors.torch import load_file
>>> import torch
```

Download safetensors & torch weights for gpt2:

```py
>>> sf_filename = hf_hub_download("gpt2", filename="model.safetensors")
>>> pt_filename = hf_hub_download("gpt2", filename="pytorch_model.bin")
```

### CPU benchmark

```py
>>> start_st = datetime.datetime.now()
>>> weights = load_file(sf_filename, device="cpu")
>>> load_time_st = datetime.datetime.now() - start_st
>>> print(f"Loaded safetensors {load_time_st}")

>>> start_pt = datetime.datetime.now()
>>> weights = torch.load(pt_filename, map_location="cpu")
>>> load_time_pt = datetime.datetime.now() - start_pt
>>> print(f"Loaded pytorch {load_time_pt}")

>>> print(f"on CPU, safetensors is faster than pytorch by: {load_time_pt/load_time_st:.1f} X")
Loaded safetensors 0:00:00.004015
Loaded pytorch 0:00:00.307460
on CPU, safetensors is faster than pytorch by: 76.6 X
```

This speedup is due to the fact that this library avoids unnecessary copies by mapping the file directly. It is actually possible to do on [pure pytorch](https://gist.github.com/Narsil/3edeec2669a5e94e4707aa0f901d2282).
The currently shown speedup was gotten on:
* OS: Ubuntu 18.04.6 LTS
* CPU: Intel(R) Xeon(R) CPU @ 2.00GHz


### GPU benchmark

```py
>>> # This is required because this feature hasn't been fully verified yet, but 
>>> # it's been tested on many different environments
>>> os.environ["SAFETENSORS_FAST_GPU"] = "1"

>>> # CUDA startup out of the measurement
>>> torch.zeros((2, 2)).cuda()

>>> start_st = datetime.datetime.now()
>>> weights = load_file(sf_filename, device="cuda:0")
>>> load_time_st = datetime.datetime.now() - start_st
>>> print(f"Loaded safetensors {load_time_st}")

>>> start_pt = datetime.datetime.now()
>>> weights = torch.load(pt_filename, map_location="cuda:0")
>>> load_time_pt = datetime.datetime.now() - start_pt
>>> print(f"Loaded pytorch {load_time_pt}")

>>> print(f"on GPU, safetensors is faster than pytorch by: {load_time_pt/load_time_st:.1f} X")
Loaded safetensors 0:00:00.165206
Loaded pytorch 0:00:00.353889
on GPU, safetensors is faster than pytorch by: 2.1 X
```

The speedup works because this library is able to skip unnecessary CPU allocations. It is unfortunately not replicable in pure pytorch as far as we know. The library works by memory mapping the file, creating the tensor empty with pytorch and calling `cudaMemcpy` directly to move the tensor directly on the GPU.
The currently shown speedup was gotten on:
* OS: Ubuntu 18.04.6 LTS.
* GPU: Tesla T4
* Driver Version: 460.32.03
* CUDA Version: 11.2

# metadata validations

    `metadata` must be a list.

---

    `metadata` must be a named list of scalar characters.

---

    `metadata` must be a named list of scalar characters.

---

    `metadata` must be a named list.

# Security Policy

## Reporting a Vulnerability

If you believe you have found a security issue in safetensors, please do **not** open a public GitHub issue.

Instead, email [security@huggingface.co](mailto:security@huggingface.co) with a description of the issue, steps to reproduce, and any relevant details. Someone from the Hugging Face security team will review your report and recommend next steps.

You may also disclose your report through [Huntr](https://huntr.com), a vulnerability disclosure program for open-source projects.

## Hugging Face Hub and remote artefacts

Safetensors is open-source software that defines a simple format for storing tensors safely, and provides fast
(zero-copy) Rust and Python libraries to read and write it. It was designed specifically to prevent the arbitrary
code execution risks associated with formats like [pickle](https://docs.python.org/3/library/pickle.html).

While safetensors can be used fully offline, it is commonly paired with the Hugging Face Hub to download model
weights uploaded by others. When consuming artefacts from any platform, you expose yourself to risks. The
recommendations below help keep your runtime and local environment safe.

### Remote artefacts

Models uploaded on the Hugging Face Hub come in different formats. We heavily recommend uploading and downloading
models in the [`safetensors`](https://github.com/huggingface/safetensors) format, which cannot execute arbitrary code when loaded.

When loading a model through a downstream library that supports multiple formats (e.g. `transformers`, `diffusers`),
prefer the option that forces the use of safetensors (such as `use_safetensors=True` in `transformers`) so that
loading fails loudly rather than silently falling back to an unsafe format.

We also recommend pinning a specific revision of the repository you download from, to protect yourself from
upstream changes to the weights.

<!-- DISABLE-FRONTMATTER-SECTIONS -->

<div class="flex justify-center">
    <img class="block dark:hidden" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-logo-light.svg"/>
    <img class="hidden dark:block" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-logo-dark.svg"/>
</div>

# Safetensors

Safetensors is a new simple format for storing tensors safely (as opposed to pickle) and that is still fast (zero-copy). Safetensors is really [fast 🚀](./speed).

## Installation

with pip:
```
pip install safetensors
```

with conda:
```
conda install -c huggingface safetensors
```

## Usage

### Load tensors

```python
from safetensors import safe_open

tensors = {}
with safe_open("model.safetensors", framework="pt", device=0) as f:
    for k in f.keys():
        tensors[k] = f.get_tensor(k)
```

Loading only part of the tensors (interesting when running on multiple GPU)

```python
from safetensors import safe_open

tensors = {}
with safe_open("model.safetensors", framework="pt", device=0) as f:
    tensor_slice = f.get_slice("embedding")
    vocab_size, hidden_dim = tensor_slice.get_shape()
    tensor = tensor_slice[:, :hidden_dim]
```

### Save tensors

```python
import torch
from safetensors.torch import save_file

tensors = {
    "embedding": torch.zeros((2, 2)),
    "attention": torch.zeros((2, 3))
}
save_file(tensors, "model.safetensors")
```

## Format

Let's say you have safetensors file named `model.safetensors`, then `model.safetensors` will have the following internal format:

<div class="flex justify-center">
    <img src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-format.svg"/>
</div>

## Featured Projects

Safetensors is being used widely at leading AI enterprises, such as [Hugging Face](https://huggingface.co/), [EleutherAI](https://www.eleuther.ai/), and [StabilityAI](https://stability.ai/). Here is a non-exhaustive list of projects that are using safetensors:

* [huggingface/transformers](https://github.com/huggingface/transformers)
* [ml-explore/mlx](https://github.com/ml-explore/mlx)
* [huggingface/candle](https://github.com/huggingface/candle)
* [AUTOMATIC1111/stable-diffusion-webui](https://github.com/AUTOMATIC1111/stable-diffusion-webui)
* [Llama-cpp](https://github.com/ggerganov/llama.cpp/blob/e6a46b0ed1884c77267dc70693183e3b7164e0e0/convert.py#L537)
* [microsoft/TaskMatrix](https://github.com/microsoft/TaskMatrix)
* [hpcaitech/ColossalAI](https://github.com/hpcaitech/ColossalAI)
* [huggingface/pytorch-image-models](https://github.com/huggingface/pytorch-image-models)
* [CivitAI](https://civitai.com/)
* [huggingface/diffusers](https://github.com/huggingface/diffusers)
* [coreylowman/dfdx](https://github.com/coreylowman/dfdx)
* [invoke-ai/InvokeAI](https://github.com/invoke-ai/InvokeAI)
* [oobabooga/text-generation-webui](https://github.com/oobabooga/text-generation-webui)
* [Sanster/lama-cleaner](https://github.com/Sanster/lama-cleaner)
* [PaddlePaddle/PaddleNLP](https://github.com/PaddlePaddle/PaddleNLP)
* [AIGC-Audio/AudioGPT](https://github.com/AIGC-Audio/AudioGPT)
* [brycedrennan/imaginAIry](https://github.com/brycedrennan/imaginAIry)
* [comfyanonymous/ComfyUI](https://github.com/comfyanonymous/ComfyUI)
* [LianjiaTech/BELLE](https://github.com/LianjiaTech/BELLE)
* [alvarobartt/safejax](https://github.com/alvarobartt/safejax)
* [MaartenGr/BERTopic](https://github.com/MaartenGr/BERTopic)

# Torch shared tensors


## TL;DR

Using specific functions, which should work in most cases for you.
This is not without side effects.

```python
from safetensors.torch import load_model, save_model

save_model(model, "model.safetensors")
# Instead of save_file(model.state_dict(), "model.safetensors")

load_model(model, "model.safetensors")
# Instead of model.load_state_dict(load_file("model.safetensors"))
```

## What are shared tensors ?

Pytorch uses shared tensors for some computation.
This is extremely interesting to reduce memory usage in general.

One very classic use case is in transformers the `embeddings` are shared with
`lm_head`. By using the same matrix, the model uses less parameters, and gradients 
flow much better to the `embeddings` (which is the start of the model, so they don't
flow easily there, whereas `lm_head` is at the tail of the model, so gradients are
extremely good over there, since they are the same tensors, they both benefit)


```python
from torch import nn

class Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.a = nn.Linear(100, 100)
        self.b = self.a

    def forward(self, x):
        return self.b(self.a(x))


model = Model()
print(model.state_dict())
# odict_keys(['a.weight', 'a.bias', 'b.weight', 'b.bias'])
torch.save(model.state_dict(), "model.bin")
# This file is now 41k instead of ~80k, because A and B are the same weight hence only 1 is saved on disk with both `a` and `b` pointing to the same buffer
```

## Why are shared tensors not saved in `safetensors` ?

Multiple reasons for that:

- *Not all frameworks support them* for instance `tensorflow` does not.
  So if someone saves shared tensors in torch, there is no way to 
  load them in a similar fashion so we could not keep the same `Dict[str, Tensor]`
  API.
- *It makes lazy loading very quickly.*
  Lazy loading is the ability to load only some tensors, or part of tensors for
  a given file. This is trivial to do without sharing tensors but with tensor sharing

  ```python
  with safe_open("model.safetensors", framework="pt") as f:
      a = f.get_tensor("a")
      b = f.get_tensor("b")
  ```

  Now it's impossible with this given code to "reshare" buffers after the fact.
  Once we give the `a` tensor we have no way to give back the same memory when
  you ask for `b`. (In this particular example we could keep track of given buffers
  but this is not the case in general, since you could do arbitrary work with `a`
  like sending it to another device before asking for `b`)
- *It can lead to much larger file than necessary*.
  If you are saving a shared tensor which is only a fraction of a larger tensor,
  then saving it with pytorch leads to saving the entire buffer instead of saving
  just what is needed.

  ```python
  a = torch.zeros((100, 100))
  b = a[:1, :]
  torch.save({"b": b}, "model.bin")
  # File is 41k instead of the expected 400 bytes
  # In practice it could happen that you save several 10GB instead of 1GB.
  ```

Now with all those reasons being mentioned, nothing is set in stone in there.
Shared tensors do not cause unsafety, or denial of service potential, so this
decision could be revisited if current workarounds are not satisfactory.

## How does it work ?

The design is rather simple.
We're going to look for all shared tensors, then looking for all tensors
covering the entire buffer (there can be multiple such tensors).
That gives us multiple names which can be saved, we simply choose the first one

During `load_model`, we are loading a bit like `load_state_dict` does, except
we're looking into the model itself, to check for shared buffers, and ignoring
the "missed keys" which were actually covered by virtue of buffer sharing (they
were properly loaded since there was a buffer that loaded under the hood).
Every other error is raised as-is

**Caveat**: This means we're dropping some keys within the file. meaning if you're
checking for the keys saved on disk, you will see some "missing tensors" or if you're
using `load_state_dict`. Unless we start supporting shared tensors directly in
the format there's no real way around it.

# Releasing safetensors

This document covers the full release process for the safetensors project. If anything here is unclear or out of date, please open a PR.

## What gets released

A single tag push triggers two releases:

- **Rust core crate** on [crates.io](https://crates.io/crates/safetensors), via `.github/workflows/rust-release.yml`
- **Python wheels + sdist** on [PyPI](https://pypi.org/project/safetensors), via `.github/workflows/python-release.yml` (with build-provenance attestations)

Both workflows trigger on tags matching `v*` (e.g. `v0.8.0`, `v0.8.0-rc.0`).

Conda packages are published separately via the [conda-forge/safetensors-feedstock](https://github.com/conda-forge/safetensors-feedstock), which picks up new PyPI releases automatically. No action is required from us.

The Rust crate and the Python binding share a version number. They could be versioned independently, but in practice they've always moved together.

## Pre-release checklist

1. **CI is green** on `main`, every platform, every Python version.
2. **Review the diff** since the last release:
   ```bash
   git log --oneline v0.7.0..main
   git diff v0.7.0..main --stat
   ```
3. **Identify breaking changes.** Anything that changes the public Rust or Python API should be called out in the release notes and reflected in the version bump per [semver](https://semver.org).
4. **Run benchmarks** if the release includes performance-sensitive work:
   ```bash
   cd safetensors && cargo bench
   ```
5. **Test against `transformers`.** It's the largest downstream consumer of safetensors.
   - At minimum, run the fast test suite:
     ```bash
     RUN_PIPELINE_TESTS=1 CUDA_VISIBLE_DEVICES=-1 pytest -sv tests/
     ```
   - For any significant release, run the full suite by rebasing these two PRs on the `transformers` repo:
     - [transformers#16708](https://github.com/huggingface/transformers/pull/16708) — builds docker images with `safetensors@main`
     - [transformers#16712](https://github.com/huggingface/transformers/pull/16712) — runs the full test suite
     If those PRs have drifted, ask the `transformers` team for the current entry point.
6. **Check the `transformers` version pin.** `transformers` pins `safetensors>=X,<Y` in its [setup.py](https://github.com/huggingface/transformers/blob/main/setup.py). If this release crosses the upper bound, coordinate with the `transformers` maintainers first.

## Cutting the release

We typically cut at least one release candidate before GA.

### 1. Update the dev version on `main`

`main` always carries a `-dev.0` version. Before branching, make sure it reflects the upcoming release:

```bash
# In both safetensors/Cargo.toml and bindings/python/Cargo.toml:
# e.g. 0.7.0-dev.0 → 0.8.0-dev.0
```

Update lockfiles accordingly:

```bash
cargo check
cd bindings/python && uv lock
```

Merge this to `main`.

### 2. Branch off `main`

```bash
git checkout main && git pull
git checkout -b git_v0.8.0
```

### 3. Set the release version on the branch

On the release branch, update both Cargo.toml files to the actual release version (drop the `-dev.0`):

```
0.8.0-dev.0 → 0.8.0
```

Update lockfiles again:

```bash
cargo check
cd bindings/python && uv lock
```

Commit:

```
Set version to 0.8.0
```

The Python version is not set in `pyproject.toml` directly — maturin reads it from `bindings/python/Cargo.toml` at build time.

### 4. Create the release on GitHub

Go to the [GitHub Releases page](https://github.com/huggingface/safetensors/releases) and draft a new release:

- **Choose the release branch** (e.g. `git_v0.8.0`)
- **Create a new tag** on publish: `v0.8.0`
- **Generate release notes** from the previous tag (e.g. `v0.7.0`)
- Add any notable highlights at the top of the generated notes

Structure for manual notes:

```markdown
## Breaking changes
- ...

## New features
- ...

## Bug fixes
- ...

## Internal / CI
- ...
```

Click **Publish release**. This creates the tag, which triggers the three release workflows.

### 5. Monitor the CI

The tag triggers two workflow runs in the [Actions tab](https://github.com/huggingface/safetensors/actions):

- **CI** (Python release) — builds wheels for every platform × Python version, uploads to PyPI. Can take up to over 30 minutes.
- **Rust Release** — `cargo publish` to crates.io. A few seconds to minutes.

The Python release uses `--skip-existing`, so re-running a partially failed workflow is safe. If `main` was green, the release builds should be too.

### 6. Verify

```bash
pip install safetensors==0.8.0
python -c "import safetensors; print(safetensors.__version__)"
```

Also check:
- [crates.io/crates/safetensors](https://crates.io/crates/safetensors)
- [pypi.org/project/safetensors](https://pypi.org/project/safetensors)

### 7. Bump `main` to the next dev version

After the release, open a PR on `main` setting both Cargo.toml files to the next dev version:

```
0.8.0-dev.0 → 0.9.0-dev.0
```

This keeps `main` clearly marked as unreleased.

## Release candidates

For releases with significant changes, cut an RC first:

- On the release branch, set the version to `0.8.0-rc.0` instead of `0.8.0`
- Tag as `v0.8.0-rc.0`, publish as a **pre-release** on GitHub
- Let it soak for a few days, ask downstream users to try it
- Fix regressions if needed (cut `-rc.1`, repeat)
- When ready, update the branch to `0.8.0` and publish the GA release

## Hotfixing a release branch

Sometimes a release or RC has a problem that needs fixing before GA (e.g. a CI matrix still referencing a dropped Python version). The fix goes on the release branch first, then gets upstreamed to `main`:

1. **Fix on the release branch.** Commit the fix directly to the release branch (e.g. `git_v0.8.0`).
2. **Re-tag and re-publish.** The procedure depends on what changed:
   - **CI-only fix** (no code changes in the published artifacts): delete the old tag, re-publish the same version. The wheels are identical so `--skip-existing` handles PyPI; only the failed jobs re-run.
     ```bash
     git tag -d v0.8.0-rc.0
     git push origin :refs/tags/v0.8.0-rc.0
     ```
     Go to the [Releases page](https://github.com/huggingface/safetensors/releases) — the previous release should appear as a draft. Re-publish it; this re-creates the tag on the latest commit and triggers the workflows.
   - **Code change**: bump the RC number in both Cargo.toml files (`0.8.0-rc.0` → `0.8.0-rc.1`). Leave the old tag in place (it's correct for its commit) and create a new release on GitHub with the new tag (`v0.8.0-rc.1`) pointing at the release branch.
3. **Cherry-pick to `main`.** Once the fix is validated on the release branch (CI green, artifacts published), cherry-pick it back to `main` via a feature branch:
   ```bash
   git checkout main && git pull
   git checkout -b fix/backport-<short-description>
   git cherry-pick <fix-commit-sha>
   git push origin fix/backport-<short-description>
   ```
   Open a PR from the feature branch so CI runs against `main` before merging. This ensures the fix isn't lost when the next release branches off.

## Testing release CI changes

If you're modifying the release workflows:

1. Comment out the upload steps (`maturin upload` / `cargo publish`).
2. Temporarily change the trigger to `push` on your branch.
3. Iterate until the artifacts build cleanly.
4. Revert both changes before merging.

## Troubleshooting

- **PyPI upload failed mid-matrix.** `--skip-existing` makes re-runs safe. Re-run from the Actions UI.
- **`cargo publish` says "already published".** You can't re-publish the same version. Bump to the next patch or pre-release and tag again.
- **Conda package not updated.** Conda packages are published via [conda-forge/safetensors-feedstock](https://github.com/conda-forge/safetensors-feedstock), which tracks PyPI. The feedstock's bot usually opens an auto-update PR within a day of the PyPI release; if it hasn't, ping the feedstock maintainers or open the PR yourself.
- **Workflow didn't trigger.** Verify the tag was pushed and that the name matches `v*`.

<!-- DISABLE-FRONTMATTER-SECTIONS -->

<div class="flex justify-center">
    <img class="block dark:hidden" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-logo-light.svg"/>
    <img class="hidden dark:block" src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-logo-dark.svg"/>
</div>

# Safetensors

Safetensors is a new simple format for storing tensors safely (as opposed to pickle) and that is still fast (zero-copy). Safetensors is really [fast 🚀](./speed).

## Installation

with pip:
```
pip install safetensors
```

with conda:
```
conda install -c conda-forge safetensors
```

## Usage

### Load tensors

```python
from safetensors import safe_open

tensors = {}
with safe_open("model.safetensors", framework="pt", device=0) as f:
    for k in f.keys():
        tensors[k] = f.get_tensor(k)
```

Loading only part of the tensors (interesting when running on multiple GPU)

```python
from safetensors import safe_open

tensors = {}
with safe_open("model.safetensors", framework="pt", device=0) as f:
    tensor_slice = f.get_slice("embedding")
    vocab_size, hidden_dim = tensor_slice.get_shape()
    tensor = tensor_slice[:, :hidden_dim]
```

### Save tensors

```python
import torch
from safetensors.torch import save_file

tensors = {
    "embedding": torch.zeros((2, 2)),
    "attention": torch.zeros((2, 3))
}
save_file(tensors, "model.safetensors")
```

## Format

Let's say you have safetensors file named `model.safetensors`, then `model.safetensors` will have the following internal format:

<div class="flex justify-center">
    <img src="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/safetensors/safetensors-format.svg"/>
</div>

## Featured Projects

Safetensors is being used widely at leading AI enterprises, such as [Hugging Face](https://huggingface.co/), [EleutherAI](https://www.eleuther.ai/), and [StabilityAI](https://stability.ai/). Here is a non-exhaustive list of projects that are using safetensors:

* [huggingface/transformers](https://github.com/huggingface/transformers)
* [ml-explore/mlx](https://github.com/ml-explore/mlx)
* [huggingface/candle](https://github.com/huggingface/candle)
* [AUTOMATIC1111/stable-diffusion-webui](https://github.com/AUTOMATIC1111/stable-diffusion-webui)
* [Llama-cpp](https://github.com/ggerganov/llama.cpp/blob/e6a46b0ed1884c77267dc70693183e3b7164e0e0/convert.py#L537)
* [microsoft/TaskMatrix](https://github.com/microsoft/TaskMatrix)
* [hpcaitech/ColossalAI](https://github.com/hpcaitech/ColossalAI)
* [huggingface/pytorch-image-models](https://github.com/huggingface/pytorch-image-models)
* [CivitAI](https://civitai.com/)
* [huggingface/diffusers](https://github.com/huggingface/diffusers)
* [coreylowman/dfdx](https://github.com/coreylowman/dfdx)
* [invoke-ai/InvokeAI](https://github.com/invoke-ai/InvokeAI)
* [oobabooga/text-generation-webui](https://github.com/oobabooga/text-generation-webui)
* [Sanster/lama-cleaner](https://github.com/Sanster/lama-cleaner)
* [PaddlePaddle/PaddleNLP](https://github.com/PaddlePaddle/PaddleNLP)
* [AIGC-Audio/AudioGPT](https://github.com/AIGC-Audio/AudioGPT)
* [brycedrennan/imaginAIry](https://github.com/brycedrennan/imaginAIry)
* [comfyanonymous/ComfyUI](https://github.com/comfyanonymous/ComfyUI)
* [LianjiaTech/BELLE](https://github.com/LianjiaTech/BELLE)
* [alvarobartt/safejax](https://github.com/alvarobartt/safejax)
* [MaartenGr/BERTopic](https://github.com/MaartenGr/BERTopic)
* [rachthree/safestructures](https://github.com/rachthree/safestructures)
* [justinchuby/onnx-safetensors](https://github.com/justinchuby/onnx-safetensors)

# Maintainers

This document lists the Maintainers of the Project. Maintainers may be added once approved by the existing maintainers as described in the [Governance document](./GOVERNANCE.md). By adding your name to this list you are agreeing to abide by the Project governance documents and to abide by all of the Organization's polices, including the [code of conduct](./CODE-OF-CONDUCT.md). If you are participating because of your affiliation with another organization (designated below), you represent that you have the authority to bind that organization to these policies.

| **NAME**      | **Organization** |
|---------------|------------------|
| Daniël de Kok | Hugging Face     |
| Luc Georges   | Hugging Face     |

---
Part of MVG-0.1-beta.
Made with love by GitHub. Licensed under the [CC-BY 4.0 License](https://creativecommons.org/licenses/by/4.0/).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safetensors.R
\name{safetensors}
\alias{safetensors}
\title{Low level control over safetensors files}
\description{
Low level control over safetensors files

Low level control over safetensors files
}
\details{
Allows opening a connection to a safetensors file and query the tensor names,
metadata, etc.
Opening a connection only reads the file metadata into memory.
This allows for more fined grained control over reading.
}
\examples{
if (rlang::is_installed("torch") && torch::torch_is_installed()) {
tensors <- list(x = torch::torch_randn(10, 10))
temp <- tempfile()
safe_save_file(tensors, temp)
f <- safetensors$new(temp, framework = "torch")
f$get_tensor("x")
}

}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{con}}{the connection object with the file}

\item{\code{metadata}}{an R list containing the metadata header in the file}

\item{\code{framework}}{the framework used to return the tensors}

\item{\code{args}}{additional arguments for tensor creation}

\item{\code{max_offset}}{the largest offset boundary that was visited. Mainly
used in torch to find the end of the safetensors file.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-safetensors-new}{\code{safetensors$new()}}
\item \href{#method-safetensors-keys}{\code{safetensors$keys()}}
\item \href{#method-safetensors-get_tensor}{\code{safetensors$get_tensor()}}
\item \href{#method-safetensors-clone}{\code{safetensors$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-safetensors-new"></a>}}
\if{latex}{\out{\hypertarget{method-safetensors-new}{}}}
\subsection{Method \code{new()}}{
Opens the connection with the file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{safetensors$new(path, ..., framework)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Path to the file to load}

\item{\code{...}}{(any)\cr
Additional, framework dependent, arguments to pass to use when creating the tensor.
For torch, this is the device, for pjrt the client.}

\item{\code{framework}}{Framework to load the data into. Currently supports "torch" and "pjrt"}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-safetensors-keys"></a>}}
\if{latex}{\out{\hypertarget{method-safetensors-keys}{}}}
\subsection{Method \code{keys()}}{
Get the keys (tensor names) in the file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{safetensors$keys()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-safetensors-get_tensor"></a>}}
\if{latex}{\out{\hypertarget{method-safetensors-get_tensor}{}}}
\subsection{Method \code{get_tensor()}}{
Get a tensor from its name
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{safetensors$get_tensor(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the tensor to load}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-safetensors-clone"></a>}}
\if{latex}{\out{\hypertarget{method-safetensors-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{safetensors$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{safe_save_file}
\alias{safe_save_file}
\alias{safe_serialize}
\title{Writes a list of tensors to the safetensors format}
\usage{
safe_save_file(tensors, path, ..., metadata = NULL)

safe_serialize(tensors, ..., metadata = NULL)
}
\arguments{
\item{tensors}{A named list of tensors. Currently only torch tensors are supported.}

\item{path}{The path to save the tensors to. It can also be a binary connection, as eg,
created with \code{file()}.}

\item{...}{Currently unused.}

\item{metadata}{An optional string that is added to the file header. Possibly
adding additional description to the weights.}
}
\value{
The path invisibly or a raw vector.
}
\description{
Writes a list of tensors to the safetensors format
}
\section{Functions}{
\itemize{
\item \code{safe_serialize()}: Serializes the tensors and returns a raw vector.

}}
\examples{
if (rlang::is_installed("torch") && torch::torch_is_installed()) {
  tensors <- list(x = torch::torch_randn(10, 10))
  temp <- tempfile()
  safe_save_file(tensors, temp)
  safe_load_file(temp, framework = "torch")

  ser <- safe_serialize(tensors)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{data}
\name{safetensors_frameworks}
\alias{safetensors_frameworks}
\title{Reflection of supported frameworks}
\format{
An object of class \code{environment} of length 1.
}
\usage{
safetensors_frameworks
}
\description{
A reflection of supported frameworks.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safetensors.R
\name{safe_load_file}
\alias{safe_load_file}
\title{Safe load a safetensors file}
\usage{
safe_load_file(path, ..., framework)
}
\arguments{
\item{path}{Path to the file to load}

\item{...}{Additional framework dependent arguments to pass to the tensor creation function.}

\item{framework}{Framework to load the data into. Currently supports "torch" and "pjrt"}
}
\value{
A list with tensors in the file. The \code{metadata} attribute can be used
to find metadata the metadata header in the file.
}
\description{
Loads an safetensors file from disk.
}
\examples{
if (rlang::is_installed("torch") && torch::torch_is_installed()) {
  tensors <- list(x = torch::torch_randn(10, 10))
  temp <- tempfile()
  safe_save_file(tensors, temp)
  safe_load_file(temp, framework = "torch")
}

}
\seealso{
\link{safetensors}, \code{\link[=safe_save_file]{safe_save_file()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{safe_tensor_buffer}
\alias{safe_tensor_buffer}
\title{Get raw buffer from a tensor}
\usage{
safe_tensor_buffer(x)
}
\arguments{
\item{x}{(any)\cr
Tensor object.}
}
\value{
(\code{raw})
}
\description{
Convert a tensor object to a raw buffer in the formated expected by safetensors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{safe_tensor_meta}
\alias{safe_tensor_meta}
\title{Get metadata from a tensor}
\usage{
safe_tensor_meta(x)
}
\arguments{
\item{x}{(any)\cr
Tensor object.}
}
\value{
(\code{list})
}
\description{
Get the metadata from a tensor.
}
